import { useEffect, useRef } from 'react';
import { Observable } from 'rxjs';

const maxPeakValue = 32_768;

interface Props {
  width: number;
  height: number;
  peaks: Observable<number[]>;
  channels: number;
  strokeStyle: string;
  fillStyle: string;
  zIndex: number;
  alpha: number;
}

// Canvas is a generic component that renders a waveform to a canvas.
//
// Properties:
//
// peaks: a 2d array of uint16s representing the peak values. Each inner array length should match logicalWidth
// strokeStyle: waveform style
// fillStyle: background style
// style: React.CSSProperties applied to canvas element
export const WaveformCanvas: React.FC<Props> = (props: Props) => {
  const canvasRef = useRef<HTMLCanvasElement>(null);

  useEffect(() => {
    const canvas = canvasRef.current;
    if (canvas == null) {
      console.error('no canvas ref available');
      return;
    }

    const ctx = canvas.getContext('2d');
    if (ctx == null) {
      console.error('no 2d context available');
      return;
    }

    ctx.strokeStyle = props.strokeStyle;
    ctx.fillStyle = props.fillStyle;
    ctx.fillRect(0, 0, canvas.width, canvas.height);

    if (props.peaks == null) {
      return;
    }

    const chanHeight = canvas.height / props.channels;

    let frameIndex = 0;
    props.peaks.forEach((peaks) => {
      for (let chanIndex = 0; chanIndex < peaks.length; chanIndex++) {
        const yOffset = chanHeight * chanIndex;
        const val = peaks[chanIndex];
        const height = Math.floor((val / maxPeakValue) * chanHeight);
        const y1 = (chanHeight - height) / 2 + yOffset;
        const y2 = y1 + height;
        ctx.beginPath();
        ctx.globalAlpha = props.alpha;
        ctx.moveTo(frameIndex, y1);
        ctx.lineTo(frameIndex, y2);
        ctx.stroke();
        ctx.globalAlpha = 1;
      }
      frameIndex++;
    });
  }, [props.peaks]);

  const canvasStyles = {
    display: 'block',
    position: 'absolute',
    width: '100%',
    height: '100%',
    zIndex: props.zIndex,
  } as React.CSSProperties;

  return (
    <>
      <canvas
        ref={canvasRef}
        width={props.width}
        height={props.height}
        style={canvasStyles}
      ></canvas>
    </>
  );
};
